import dataclasses as dc
import typing

import numpy as np

from pydrake.all import (
    Adder,
    AddMultibodyPlant,
    ApplyLcmBusConfig,
    ApplyMultibodyPlantConfig,
    ApplyVisualizationConfig,
    CameraConfig,
    Demultiplexer,
    DiagramBuilder,
    DrakeLcmParams,
    IiwaCommandSender,
    IiwaDriver,
    IiwaStatusReceiver,
    InverseDynamicsController,
    LcmPublisherSystem,
    LcmSubscriberSystem,
    MakeMultibodyStateToWsgStateSystem,
    Meshcat,
    ModelDirective,
    ModelDirectives,
    MultibodyPlant,
    MultibodyPlantConfig,
    Parser,
    PassThrough,
    ProcessModelDirectives,
    SceneGraph,
    SchunkWsgDriver,
    SchunkWsgPositionController,
    SchunkWsgCommandSender,
    SchunkWsgStatusReceiver,
    SimulatorConfig,
    StateInterpolatorWithDiscreteDerivative,
    VisualizationConfig,
    ZeroForceDriver,
)
from pydrake.common.yaml import yaml_load_typed
from drake import (
    lcmt_iiwa_command,
    lcmt_iiwa_status,
    lcmt_schunk_wsg_command,
    lcmt_schunk_wsg_status,
)

from manipulation.scenarios import (
    AddIiwa,
    AddPlanarIiwa,
    AddWsg,
)
from manipulation.utils import ConfigureParser


@dc.dataclass
class Scenario:
    """Defines the YAML format for a (possibly stochastic) scenario to be
    simulated.
    """

    # Random seed for any random elements in the scenario. The seed is always
    # deterministic in the `Scenario`; a caller who wants randomness must
    # populate this value from their own randomness.
    random_seed: int = 0

    # The maximum simulation time (in seconds).  The simulator will attempt to
    # run until this time and then terminate.
    simulation_duration: float = np.inf

    # Simulator configuration (integrator and publisher parameters).
    simulator_config: SimulatorConfig = SimulatorConfig(
        max_step_size=0.01,
        use_error_control=False,
        accuracy=1.0e-2,
    )

    # Plant configuration (time step and contact parameters).
    plant_config: MultibodyPlantConfig = MultibodyPlantConfig(
        discrete_contact_solver="sap"
    )

    # All of the fully deterministic elements of the simulation.
    directives: typing.List[ModelDirective] = dc.field(default_factory=list)

    # A map of {bus_name: lcm_params} for LCM transceivers to be used by
    # drivers, sensors, etc.
    lcm_buses: typing.Mapping[str, DrakeLcmParams] = dc.field(
        default_factory=lambda: dict(default=DrakeLcmParams())
    )

    # For actuated models, specifies where each model's actuation inputs come
    # from, keyed on the ModelInstance name.
    model_drivers: typing.Mapping[
        str,
        typing.Union[
            IiwaDriver,
            SchunkWsgDriver,
            ZeroForceDriver,
        ],
    ] = dc.field(default_factory=dict)

    # Cameras to add to the scene (and broadcast over LCM). The key for each
    # camera is a helpful mnemonic, but does not serve a technical role. The
    # CameraConfig::name field is still the name that will appear in the
    # Diagram artifacts.
    cameras: typing.Mapping[str, CameraConfig] = dc.field(default_factory=dict)

    visualization: VisualizationConfig = VisualizationConfig()


def load_scenario(*, filename=None, data=None, scenario_name=None):
    """Implements the command-line handling logic for scenario data.
    Returns a `Scenario` object loaded from the given input arguments.

    Args:
        filename (optional): A yaml filename to load the scenario from.

        data (optional): A yaml string to load the scenario from. If both
            filename and string are specified, then the filename is parsed
            first, and then the string is _also_ parsed, potentially overwriting defaults from the filename..

        scenario_name (optional): The name of the scenario/child to load from
            the yaml file. If None, then the entire file is loaded.
    """
    result = Scenario()
    if filename:
        result = yaml_load_typed(
            schema=Scenario,
            filename=filename,
            child_name=scenario_name,
            defaults=result,
        )
    if data:
        result = yaml_load_typed(schema=Scenario, data=data, defaults=result)
    return result


# TODO(russt): Use the c++ version pending https://github.com/RobotLocomotion/drake/issues/20055
def ApplyDriverConfigSim(
    driver_config,
    model_instance_name,
    sim_plant,
    models_from_directives_map,
    builder,
):
    model_instance = sim_plant.GetModelInstanceByName(model_instance_name)
    if isinstance(driver_config, IiwaDriver):
        num_iiwa_positions = sim_plant.num_positions(model_instance)

        # I need a PassThrough system so that I can export the input port.
        iiwa_position = builder.AddSystem(PassThrough(num_iiwa_positions))
        builder.ExportInput(
            iiwa_position.get_input_port(),
            model_instance_name + ".position",
        )
        builder.ExportOutput(
            iiwa_position.get_output_port(),
            model_instance_name + ".position_commanded",
        )

        # Export the iiwa "state" outputs.
        demux = builder.AddSystem(
            Demultiplexer(2 * num_iiwa_positions, num_iiwa_positions)
        )
        builder.Connect(
            sim_plant.get_state_output_port(model_instance),
            demux.get_input_port(),
        )
        builder.ExportOutput(
            demux.get_output_port(0),
            model_instance_name + ".position_measured",
        )
        builder.ExportOutput(
            demux.get_output_port(1),
            model_instance_name + ".velocity_estimated",
        )
        builder.ExportOutput(
            sim_plant.get_state_output_port(model_instance),
            model_instance_name + ".state_estimated",
        )

        # Make the plant for the iiwa controller to use.
        controller_plant = MultibodyPlant(time_step=sim_plant.time_step())
        # TODO: Add the correct IIWA model (introspected from MBP). See
        # build_iiwa_control in Drake for a slightly closer attempt.
        if num_iiwa_positions == 3:
            controller_iiwa = AddPlanarIiwa(controller_plant)
        else:
            controller_iiwa = AddIiwa(controller_plant)
        AddWsg(controller_plant, controller_iiwa, welded=True)
        controller_plant.Finalize()

        # Add the iiwa controller
        iiwa_controller = builder.AddSystem(
            InverseDynamicsController(
                controller_plant,
                kp=[100] * num_iiwa_positions,
                ki=[1] * num_iiwa_positions,
                kd=[20] * num_iiwa_positions,
                has_reference_acceleration=False,
            )
        )
        iiwa_controller.set_name(model_instance_name + ".controller")
        builder.Connect(
            sim_plant.get_state_output_port(model_instance),
            iiwa_controller.get_input_port_estimated_state(),
        )

        # Add in the feed-forward torque
        adder = builder.AddSystem(Adder(2, num_iiwa_positions))
        builder.Connect(
            iiwa_controller.get_output_port_control(),
            adder.get_input_port(0),
        )
        # Use a PassThrough to make the port optional (it will provide zero
        # values if not connected).
        torque_passthrough = builder.AddSystem(
            PassThrough([0] * num_iiwa_positions)
        )
        builder.Connect(
            torque_passthrough.get_output_port(), adder.get_input_port(1)
        )
        builder.ExportInput(
            torque_passthrough.get_input_port(),
            model_instance_name + ".feedforward_torque",
        )
        builder.Connect(
            adder.get_output_port(),
            sim_plant.get_actuation_input_port(model_instance),
        )

        # Add discrete derivative to command velocities.
        desired_state_from_position = builder.AddSystem(
            StateInterpolatorWithDiscreteDerivative(
                num_iiwa_positions,
                sim_plant.time_step(),
                suppress_initial_transient=True,
            )
        )
        desired_state_from_position.set_name(
            model_instance_name + ".desired_state_from_position"
        )
        builder.Connect(
            desired_state_from_position.get_output_port(),
            iiwa_controller.get_input_port_desired_state(),
        )
        builder.Connect(
            iiwa_position.get_output_port(),
            desired_state_from_position.get_input_port(),
        )

        # Export commanded torques.
        builder.ExportOutput(
            adder.get_output_port(),
            model_instance_name + ".torque_commanded",
        )
        builder.ExportOutput(
            adder.get_output_port(),
            model_instance_name + ".torque_measured",
        )

        builder.ExportOutput(
            sim_plant.get_generalized_contact_forces_output_port(
                model_instance
            ),
            model_instance_name + ".torque_external",
        )

    if isinstance(driver_config, SchunkWsgDriver):
        # Wsg controller.
        wsg_controller = builder.AddSystem(SchunkWsgPositionController())
        wsg_controller.set_name(model_instance_name + ".controller")
        builder.Connect(
            wsg_controller.get_generalized_force_output_port(),
            sim_plant.get_actuation_input_port(model_instance),
        )
        builder.Connect(
            sim_plant.get_state_output_port(model_instance),
            wsg_controller.get_state_input_port(),
        )
        builder.ExportInput(
            wsg_controller.get_desired_position_input_port(),
            model_instance_name + ".position",
        )
        builder.ExportInput(
            wsg_controller.get_force_limit_input_port(),
            model_instance_name + ".force_limit",
        )
        wsg_mbp_state_to_wsg_state = builder.AddSystem(
            MakeMultibodyStateToWsgStateSystem()
        )
        builder.Connect(
            sim_plant.get_state_output_port(model_instance),
            wsg_mbp_state_to_wsg_state.get_input_port(),
        )
        builder.ExportOutput(
            wsg_mbp_state_to_wsg_state.get_output_port(),
            model_instance_name + ".state_measured",
        )
        builder.ExportOutput(
            wsg_controller.get_grip_force_output_port(),
            model_instance_name + ".force_measured",
        )


def ApplyDriverConfigsSim(
    *, driver_configs, sim_plant, models_from_directives, builder
):
    models_from_directives_map = dict(
        [(info.model_name, info) for info in models_from_directives]
    )
    for model_instance_name, driver_config in driver_configs.items():
        ApplyDriverConfigSim(
            driver_config,
            model_instance_name,
            sim_plant,
            models_from_directives_map,
            builder,
        )


def MakeHardwareStation(
    scenario: Scenario,
    meshcat: Meshcat = None,
    *,
    package_xmls: typing.List[str] = [],
    hardware: bool = False,
):
    """
    If `hardware=False`, (the default) returns a HardwareStation diagram containing:
      - A MultibodyPlant with populated via the directives in `scenario`.
      - A SceneGraph
      - The default Drake visualizers
      - Any robot / sensors drivers specified in the YAML description.

    If `hardware=True`, returns a HardwareStationInterface diagram containing the network interfaces to communicate directly with the hardware drivers.

    Args:
        scenario: A Scenario structure, populated using the `load_scenario` method.

        meshcat: If not None, then AddDefaultVisualization will be added to the subdiagram using this meshcat instance.

        package_xmls: A list of package.xml file paths that will be passed to the parser, using Parser.AddPackageXml().
    """
    if hardware:
        return MakeHardwareStationInterface(
            scenario=scenario, meshcat=meshcat, package_xmls=package_xmls
        )

    builder = DiagramBuilder()

    # Create the multibody plant and scene graph.
    sim_plant, scene_graph = AddMultibodyPlant(
        config=scenario.plant_config, builder=builder
    )

    parser = Parser(sim_plant)
    for p in package_xmls:
        parser.package_map().AddPackageXml(p)
    ConfigureParser(parser)

    # Add model directives.
    added_models = ProcessModelDirectives(
        directives=ModelDirectives(directives=scenario.directives),
        parser=parser,
    )

    # Now the plant is complete.
    sim_plant.Finalize()

    # Add drivers.
    ApplyDriverConfigsSim(
        driver_configs=scenario.model_drivers,
        sim_plant=sim_plant,
        models_from_directives=added_models,
        builder=builder,
    )

    # TODO(russt): Add scene cameras. https://github.com/RobotLocomotion/drake/issues/20055

    # Add visualization.
    ApplyVisualizationConfig(scenario.visualization, builder, meshcat=meshcat)

    # Export "cheat" ports.
    builder.ExportOutput(scene_graph.get_query_output_port(), "query_object")
    builder.ExportOutput(
        sim_plant.get_contact_results_output_port(), "contact_results"
    )
    builder.ExportOutput(
        sim_plant.get_state_output_port(), "plant_continuous_state"
    )
    builder.ExportOutput(sim_plant.get_body_poses_output_port(), "body_poses")

    diagram = builder.Build()
    diagram.set_name("HardwareStation")
    return diagram


# TODO(russt): Use the c++ version pending https://github.com/RobotLocomotion/drake/issues/20055
def ApplyDriverConfigInterface(
    driver_config,
    model_instance_name,
    plant,
    models_from_directives_map,
    lcm_buses,
    builder,
):
    if isinstance(driver_config, IiwaDriver):
        lcm = lcm_buses.Find(
            "Driver for " + model_instance_name, driver_config.lcm_bus
        )

        # Publish IIWA command.
        iiwa_command_sender = builder.AddSystem(IiwaCommandSender())
        # Note on publish period: IIWA driver won't respond faster than 200Hz
        iiwa_command_publisher = builder.AddSystem(
            LcmPublisherSystem.Make(
                channel="IIWA_COMMAND",
                lcm_type=lcmt_iiwa_command,
                lcm=lcm,
                publish_period=0.005,
                use_cpp_serializer=True,
            )
        )
        iiwa_command_publisher.set_name(
            model_instance_name + ".command_publisher"
        )
        builder.ExportInput(
            iiwa_command_sender.get_position_input_port(),
            model_instance_name + ".position",
        )
        builder.ExportInput(
            iiwa_command_sender.get_torque_input_port(),
            model_instance_name + ".feedforward_torque",
        )
        builder.Connect(
            iiwa_command_sender.get_output_port(),
            iiwa_command_publisher.get_input_port(),
        )
        # Receive IIWA status and populate the output ports.
        iiwa_status_receiver = builder.AddSystem(IiwaStatusReceiver())
        iiwa_status_subscriber = builder.AddSystem(
            LcmSubscriberSystem.Make(
                channel="IIWA_STATUS",
                lcm_type=lcmt_iiwa_status,
                lcm=lcm,
                use_cpp_serializer=True,
                wait_for_message_on_initialization_timeout=10,
            )
        )
        iiwa_status_subscriber.set_name(
            model_instance_name + ".status_subscriber"
        )

        # builder.Connect(
        #    iiwa_status_receiver.get_position_measured_output_port(),
        #    to_pose.get_input_port(),
        # )

        builder.ExportOutput(
            iiwa_status_receiver.get_position_commanded_output_port(),
            model_instance_name + ".position_commanded",
        )
        builder.ExportOutput(
            iiwa_status_receiver.get_position_measured_output_port(),
            model_instance_name + ".position_measured",
        )
        builder.ExportOutput(
            iiwa_status_receiver.get_velocity_estimated_output_port(),
            model_instance_name + ".velocity_estimated",
        )
        builder.ExportOutput(
            iiwa_status_receiver.get_torque_commanded_output_port(),
            model_instance_name + ".torque_commanded",
        )
        builder.ExportOutput(
            iiwa_status_receiver.get_torque_measured_output_port(),
            model_instance_name + ".torque_measured",
        )
        builder.ExportOutput(
            iiwa_status_receiver.get_torque_external_output_port(),
            model_instance_name + ".torque_external",
        )
        builder.Connect(
            iiwa_status_subscriber.get_output_port(),
            iiwa_status_receiver.get_input_port(),
        )
    if isinstance(driver_config, SchunkWsgDriver):
        lcm = lcm_buses.Find(
            "Driver for " + model_instance_name, driver_config.lcm_bus
        )

        # Publish WSG command.
        wsg_command_sender = builder.AddSystem(SchunkWsgCommandSender())
        wsg_command_publisher = builder.AddSystem(
            LcmPublisherSystem.Make(
                channel="SCHUNK_WSG_COMMAND",
                lcm_type=lcmt_schunk_wsg_command,
                lcm=lcm,
                publish_period=0.05,  # Schunk driver won't respond faster than 20Hz
                use_cpp_serializer=True,
            )
        )
        wsg_command_publisher.set_name(
            model_instance_name + ".command_publisher"
        )
        builder.ExportInput(
            wsg_command_sender.get_position_input_port(),
            model_instance_name + ".position",
        )
        builder.ExportInput(
            wsg_command_sender.get_force_limit_input_port(),
            model_instance_name + ".force_limit",
        )
        builder.Connect(
            wsg_command_sender.get_output_port(0),
            wsg_command_publisher.get_input_port(),
        )

        # Receive WSG status and populate the output ports.
        wsg_status_receiver = builder.AddSystem(SchunkWsgStatusReceiver())
        wsg_status_subscriber = builder.AddSystem(
            LcmSubscriberSystem.Make(
                channel="SCHUNK_WSG_STATUS",
                lcm_type=lcmt_schunk_wsg_status,
                lcm=lcm,
                use_cpp_serializer=True,
                wait_for_message_on_initialization_timeout=10,
            )
        )
        wsg_status_subscriber.set_name(
            model_instance_name + ".status_subscriber"
        )
        builder.ExportOutput(
            wsg_status_receiver.get_state_output_port(),
            model_instance_name + ".state_measured",
        )
        builder.ExportOutput(
            wsg_status_receiver.get_force_output_port(),
            model_instance_name + ".force_measured",
        )
        builder.Connect(
            wsg_status_subscriber.get_output_port(),
            wsg_status_receiver.get_input_port(0),
        )


def ApplyDriverConfigsInterface(
    *, driver_configs, plant, models_from_directives, lcm_buses, builder
):
    models_from_directives_map = dict(
        [(info.model_name, info) for info in models_from_directives]
    )
    for model_instance_name, driver_config in driver_configs.items():
        ApplyDriverConfigInterface(
            driver_config,
            model_instance_name,
            plant,
            models_from_directives_map,
            lcm_buses,
            builder,
        )


def MakeHardwareStationInterface(
    scenario: Scenario,
    meshcat: Meshcat = None,
    *,
    package_xmls: typing.List[str] = [],
):
    builder = DiagramBuilder()

    # Visualization
    scene_graph = builder.AddNamedSystem("scene_graph", SceneGraph())
    plant = MultibodyPlant(time_step=scenario.plant_config.time_step)
    ApplyMultibodyPlantConfig(scenario.plant_config, plant)
    plant.RegisterAsSourceForSceneGraph(scene_graph)
    parser = Parser(plant)
    for p in package_xmls:
        parser.package_map().AddPackageXml(p)
    ConfigureParser(parser)

    # Add model directives.
    added_models = ProcessModelDirectives(
        directives=ModelDirectives(directives=scenario.directives),
        parser=parser,
    )

    # Now the plant is complete.
    plant.Finalize()

    # to_pose = builder.AddSystem(MultibodyPositionToGeometryPose(plant))
    # builder.Connect(
    #     to_pose.get_output_port(),
    #     scene_graph.get_source_pose_port(plant.get_source_id()),
    # )

    # config = VisualizationConfig()
    # config.publish_contacts = False
    # config.publish_inertia = False
    # ApplyVisualizationConfig(
    #     config, builder=builder, plant=plant, meshcat=meshcat
    # )

    # Add LCM buses. (The simulator will handle polling the network for new
    # messages and dispatching them to the receivers, i.e., "pump" the bus.)
    lcm_buses = ApplyLcmBusConfig(
        lcm_buses=scenario.lcm_buses, builder=builder
    )

    # Add drivers.
    ApplyDriverConfigsInterface(
        driver_configs=scenario.model_drivers,
        plant=plant,
        models_from_directives=added_models,
        lcm_buses=lcm_buses,
        builder=builder,
    )

    # Add cameras.

    diagram = builder.Build()
    diagram.set_name("HardwareStationInterface")
    return diagram



import time
from collections import namedtuple
from functools import partial

import numpy as np
from pydrake.common.value import AbstractValue
from pydrake.geometry import (
    Cylinder,
    Rgba,
    Sphere,
)
from pydrake.math import RigidTransform, RollPitchYaw, RotationMatrix
from pydrake.solvers import BoundingBoxConstraint
from pydrake.systems.framework import (
    EventStatus,
    LeafSystem,
)

from manipulation import running_as_notebook

# Some GUI code that will be moved into Drake.


class MeshcatSliders(LeafSystem):
    """
    A system that outputs the ``value``s from meshcat sliders.

    .. pydrake_system::

      name: MeshcatSliderSystem
      output_ports:
      - slider_group_0
      - ...
      - slider_group_{N-1}
    """

    def __init__(self, meshcat, slider_names):
        """
        An output port is created for each element in the list `slider_names`.
        Each element of `slider_names` must itself be an iterable collection
        (list, tuple, set, ...) of strings, with the names of sliders that have
        *already* been added to Meshcat via Meshcat.AddSlider().

        The same slider may be used in multiple ports.
        """
        LeafSystem.__init__(self)

        self._meshcat = meshcat
        self._sliders = slider_names
        for i, slider_iterable in enumerate(self._sliders):
            port = self.DeclareVectorOutputPort(
                f"slider_group_{i}",
                len(slider_iterable),
                partial(self.DoCalcOutput, port_index=i),
            )
            port.disable_caching_by_default()

    def DoCalcOutput(self, context, output, port_index):
        for i, slider in enumerate(self._sliders[port_index]):
            output[i] = self._meshcat.GetSliderValue(slider)


class MeshcatPoseSliders(LeafSystem):
    """
    Provides a set of ipywidget sliders (to be used in a Jupyter notebook) with
    one slider for each of roll, pitch, yaw, x, y, and z.  This can be used,
    for instance, as an interface to teleoperate the end-effector of a robot.

    .. pydrake_system::

        name: PoseSliders
        input_ports:
        - pose (optional)
        output_ports:
        - pose

    The optional `pose` input port is used ONLY at initialization; it can be
    used to set the initial pose e.g. from the current pose of a MultibodyPlant
    frame.
    """

    # TODO(russt): Use namedtuple defaults parameter once we are Python >= 3.7.
    Visible = namedtuple("Visible", ("roll", "pitch", "yaw", "x", "y", "z"))
    Visible.__new__.__defaults__ = (True, True, True, True, True, True)
    MinRange = namedtuple("MinRange", ("roll", "pitch", "yaw", "x", "y", "z"))
    MinRange.__new__.__defaults__ = (-np.pi, -np.pi, -np.pi, -1.0, -1.0, -1.0)
    MaxRange = namedtuple("MaxRange", ("roll", "pitch", "yaw", "x", "y", "z"))
    MaxRange.__new__.__defaults__ = (np.pi, np.pi, np.pi, 1.0, 1.0, 1.0)
    Value = namedtuple("Value", ("roll", "pitch", "yaw", "x", "y", "z"))
    Value.__new__.__defaults__ = (0.0, 0.0, 0.0, 0.0, 0.0, 0.0)
    DecrementKey = namedtuple(
        "DecrementKey", ("roll", "pitch", "yaw", "x", "y", "z")
    )
    DecrementKey.__new__.__defaults__ = (
        "KeyQ",
        "KeyW",
        "KeyA",
        "KeyJ",
        "KeyI",
        "KeyO",
    )
    IncrementKey = namedtuple(
        "IncrementKey", ("roll", "pitch", "yaw", "x", "y", "z")
    )
    IncrementKey.__new__.__defaults__ = (
        "KeyE",
        "KeyS",
        "KeyD",
        "KeyL",
        "KeyK",
        "KeyU",
    )

    def __init__(
        self,
        meshcat,
        visible=Visible(),
        min_range=MinRange(),
        max_range=MaxRange(),
        value=Value(),
        decrement_keycode=DecrementKey(),
        increment_keycode=IncrementKey(),
        body_index=None,
    ):
        """
        Args:
            meshcat: A Meshcat instance.
            visible: An object with boolean elements for 'roll', 'pitch',
                     'yaw', 'x', 'y', 'z'; the intention is for this to be the
                     PoseSliders.Visible() namedtuple.  Defaults to all true.
            min_range, max_range, value: Objects with float values for 'roll',
                      'pitch', 'yaw', 'x', 'y', 'z'; the intention is for the
                      caller to use the PoseSliders.MinRange, MaxRange, and
                      Value namedtuples.  See those tuples for default values.
            body_index: if the body_poses input port is connected, then this
                        index determine which pose is used to set the initial
                        slider positions during the Initialization event.
        """
        LeafSystem.__init__(self)
        port = self.DeclareAbstractOutputPort(
            "pose",
            lambda: AbstractValue.Make(RigidTransform()),
            self.DoCalcOutput,
        )

        self.DeclareAbstractInputPort(
            "body_poses", AbstractValue.Make([RigidTransform()])
        )
        self.DeclareInitializationDiscreteUpdateEvent(self.Initialize)

        # The widgets themselves have undeclared state.  For now, we accept it,
        # and simply disable caching on the output port.
        # TODO(russt): consider implementing the more elaborate methods seen
        # in, e.g., LcmMessageSubscriber.
        port.disable_caching_by_default()

        self._meshcat = meshcat
        self._visible = visible
        self._value = list(value)
        self._body_index = body_index

        print("Keyboard Controls:")
        for i in range(6):
            if visible[i]:
                meshcat.AddSlider(
                    min=min_range[i],
                    max=max_range[i],
                    value=value[i],
                    step=0.02,
                    name=value._fields[i],
                    decrement_keycode=decrement_keycode[i],
                    increment_keycode=increment_keycode[i],
                )
                print(
                    f"{value._fields[i]} : {decrement_keycode[i]} / {increment_keycode[i]}"  # noqa
                )

    def __del__(self):
        for s in ["roll", "pitch", "yaw", "x", "y", "z"]:
            if visible[s]:
                self._meshcat.DeleteSlider(s)

    def SetPose(self, pose):
        """
        Sets the current value of the sliders.

        Args:
            pose: Any viable argument for the RigidTransform
                  constructor.
        """
        tf = RigidTransform(pose)
        self.SetRpy(RollPitchYaw(tf.rotation()))
        self.SetXyz(tf.translation())

    def SetRpy(self, rpy):
        """
        Sets the current value of the sliders for roll, pitch, and yaw.

        Args:
            rpy: An instance of drake.math.RollPitchYaw
        """
        self._value[0] = rpy.roll_angle()
        self._value[1] = rpy.pitch_angle()
        self._value[2] = rpy.yaw_angle()
        for i in range(3):
            if self._visible[i]:
                self._meshcat.SetSliderValue(
                    self._visible._fields[i], self._value[i]
                )

    def SetXyz(self, xyz):
        """
        Sets the current value of the sliders for x, y, and z.

        Args:
            xyz: A 3 element iterable object with x, y, z.
        """
        self._value[3:] = xyz
        for i in range(3, 6):
            if self._visible[i]:
                self._meshcat.SetSliderValue(
                    self._visible._fields[i], self._value[i]
                )

    def _update_values(self):
        changed = False
        for i in range(6):
            if self._visible[i]:
                old_value = self._value[i]
                self._value[i] = self._meshcat.GetSliderValue(
                    self._visible._fields[i]
                )
                changed = changed or self._value[i] != old_value
        return changed

    def _get_transform(self):
        return RigidTransform(
            RollPitchYaw(self._value[0], self._value[1], self._value[2]),
            self._value[3:],
        )

    def DoCalcOutput(self, context, output):
        """Constructs the output values from the sliders."""
        self._update_values()
        output.set_value(self._get_transform())

    def Initialize(self, context, discrete_state):
        if self.get_input_port().HasValue(context):
            if self._body_index is None:
                raise RuntimeError(
                    "If the `body_poses` input port is connected, then you "
                    "must also pass a `body_index` to the constructor."
                )
            self.SetPose(self.get_input_port().Eval(context)[self._body_index])
            return EventStatus.Succeeded()
        return EventStatus.DidNothing()

    def Run(self, publishing_system, root_context, callback):
        # Calls callback(root_context, pose), then
        # publishing_system.ForcedPublish() each time the sliders change value.
        if not running_as_notebook:
            return

        publishing_context = publishing_system.GetMyContextFromRoot(
            root_context
        )

        print("Press the 'Stop PoseSliders' button in Meshcat to continue.")
        self._meshcat.AddButton("Stop PoseSliders", "Escape")
        while self._meshcat.GetButtonClicks("Stop PoseSliders") < 1:
            if self._update_values():
                callback(root_context, self._get_transform())
                publishing_system.ForcedPublish(publishing_context)
            time.sleep(0.1)

        self._meshcat.DeleteButton("Stop PoseSliders")


class WsgButton(LeafSystem):
    def __init__(self, meshcat):
        LeafSystem.__init__(self)
        port = self.DeclareVectorOutputPort(
            "wsg_position", 1, self.DoCalcOutput
        )
        port.disable_caching_by_default()
        self._meshcat = meshcat
        self._button = "Open/Close Gripper"
        meshcat.AddButton(self._button, "Space")
        print("Press Space to open/close the gripper")

    def __del__(self):
        self._meshcat.DeleteButton(self._button)

    def DoCalcOutput(self, context, output):
        position = 0.107  # open
        if (self._meshcat.GetButtonClicks(self._button) % 2) == 1:
            position = 0.002  # close
        output.SetAtIndex(0, position)


def AddMeshcatTriad(
    meshcat, path, length=0.25, radius=0.01, opacity=1.0, X_PT=RigidTransform()
):
    meshcat.SetTransform(path, X_PT)
    # x-axis
    X_TG = RigidTransform(
        RotationMatrix.MakeYRotation(np.pi / 2), [length / 2.0, 0, 0]
    )
    meshcat.SetTransform(path + "/x-axis", X_TG)
    meshcat.SetObject(
        path + "/x-axis", Cylinder(radius, length), Rgba(1, 0, 0, opacity)
    )

    # y-axis
    X_TG = RigidTransform(
        RotationMatrix.MakeXRotation(np.pi / 2), [0, length / 2.0, 0]
    )
    meshcat.SetTransform(path + "/y-axis", X_TG)
    meshcat.SetObject(
        path + "/y-axis", Cylinder(radius, length), Rgba(0, 1, 0, opacity)
    )

    # z-axis
    X_TG = RigidTransform([0, 0, length / 2.0])
    meshcat.SetTransform(path + "/z-axis", X_TG)
    meshcat.SetObject(
        path + "/z-axis", Cylinder(radius, length), Rgba(0, 0, 1, opacity)
    )


def plot_surface(
    meshcat,
    path,
    X,
    Y,
    Z,
    rgba=Rgba(0.87, 0.6, 0.6, 1.0),
    wireframe=False,
    wireframe_line_width=1.0,
):
    (rows, cols) = Z.shape
    assert np.array_equal(X.shape, Y.shape)
    assert np.array_equal(X.shape, Z.shape)

    vertices = np.empty((rows * cols, 3), dtype=np.float32)
    vertices[:, 0] = X.reshape((-1))
    vertices[:, 1] = Y.reshape((-1))
    vertices[:, 2] = Z.reshape((-1))

    # Vectorized faces code from https://stackoverflow.com/questions/44934631/making-grid-triangular-mesh-quickly-with-numpy  # noqa
    faces = np.empty((rows - 1, cols - 1, 2, 3), dtype=np.uint32)
    r = np.arange(rows * cols).reshape(rows, cols)
    faces[:, :, 0, 0] = r[:-1, :-1]
    faces[:, :, 1, 0] = r[:-1, 1:]
    faces[:, :, 0, 1] = r[:-1, 1:]
    faces[:, :, 1, 1] = r[1:, 1:]
    faces[:, :, :, 2] = r[1:, :-1, None]
    faces.shape = (-1, 3)

    # TODO(Russ): support per vertex / Colormap colors.
    meshcat.SetTriangleMesh(
        path, vertices.T, faces.T, rgba, wireframe, wireframe_line_width
    )


def plot_mathematical_program(
    meshcat, path, prog, X, Y, result=None, point_size=0.05
):
    assert prog.num_vars() == 2
    assert X.size == Y.size

    X.size
    values = np.vstack((X.reshape(-1), Y.reshape(-1)))
    costs = prog.GetAllCosts()

    # Vectorized multiply for the quadratic form.
    # Z = (D*np.matmul(Q,D)).sum(0).reshape(nx, ny)

    if costs:
        Z = prog.EvalBindingVectorized(costs[0], values)
        for b in costs[1:]:
            Z = Z + prog.EvalBindingVectorized(b, values)

    cv = f"{path}/constraints"
    for binding in prog.GetAllConstraints():
        if isinstance(binding.evaluator(), BoundingBoxConstraint):
            c = binding.evaluator()
            var_indices = [
                int(prog.decision_variable_index()[v.get_id()])
                for v in binding.variables()
            ]
            satisfied = np.array(
                c.CheckSatisfiedVectorized(values[var_indices, :], 0.001)
            ).reshape(1, -1)
            if costs:
                Z[~satisfied] = np.nan

            v = f"{cv}/{type(c).__name__}"
            Zc = np.zeros(Z.shape)
            Zc[satisfied] = np.nan
            plot_surface(
                meshcat,
                v,
                X,
                Y,
                Zc.reshape(X.shape),
                rgba=Rgba(1.0, 0.2, 0.2, 1.0),
                wireframe=True,
            )
        else:
            Zc = prog.EvalBindingVectorized(binding, values)
            evaluator = binding.evaluator()
            low = evaluator.lower_bound()
            up = evaluator.upper_bound()
            cvb = f"{cv}/{type(evaluator).__name__}"
            for index in range(Zc.shape[0]):
                # TODO(russt): Plot infeasible points in a different color.
                infeasible = np.logical_or(
                    Zc[index, :] < low[index], Zc[index, :] > up[index]
                )
                plot_surface(
                    meshcat,
                    f"{cvb}/{index}",
                    X,
                    Y,
                    Zc[index, :].reshape(X.shape),
                    rgba=Rgba(1.0, 0.3, 1.0, 1.0),
                    wireframe=True,
                )

    if costs:
        plot_surface(
            meshcat,
            f"{path}/objective",
            X,
            Y,
            Z.reshape(X.shape),
            rgba=Rgba(0.3, 1.0, 0.3, 1.0),
            wireframe=True,
        )

    if result:
        v = f"{path}/solution"
        meshcat.SetObject(v, Sphere(point_size), Rgba(0.3, 1.0, 0.3, 1.0))
        x_solution = result.get_x_val()
        meshcat.SetTransform(
            v,
            RigidTransform(
                [x_solution[0], x_solution[1], result.get_optimal_cost()]
            ),
        )


def PublishPositionTrajectory(
    trajectory, root_context, plant, visualizer, time_step=1.0 / 33.0
):
    """
    Args:
        trajectory: A Trajectory instance.
    """
    plant_context = plant.GetMyContextFromRoot(root_context)
    visualizer_context = visualizer.GetMyContextFromRoot(root_context)

    visualizer.StartRecording(False)

    for t in np.append(
        np.arange(trajectory.start_time(), trajectory.end_time(), time_step),
        trajectory.end_time(),
    ):
        root_context.SetTime(t)
        plant.SetPositions(plant_context, trajectory.value(t))
        visualizer.ForcedPublish(visualizer_context)

    visualizer.StopRecording()
    visualizer.PublishRecording()



import unittest
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight
import numpy as np


class TestAnalyticGrasp(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(4)
    @timeout_decorator.timeout(10.0)
    def test_antipodal_points(self):
        """Test find_antipodal_pts"""
        shape = self.notebook_locals["shape"]
        f = self.notebook_locals["find_antipodal_pts"]

        # Use a seed that can test for different Hessians
        np.random.seed(45)

        result_lst = []
        H_eig_lst = []

        for i in range(4):
            result, H_eig = f(shape)
            result_lst.append(result)
            H_eig_lst.append(np.sort(H_eig)[::-1])

        result_lst_eval = np.array(result_lst)
        H_eig_lst_eval = np.array(H_eig_lst)

        # yapf: disable
        result_lst_target = np.array([  # noqa
            [6.276351, 3.497493],
            [3.948776, 1.636278],
            [6.104002, 4.219046],
            [3.223685, 0.710411]])

        H_eig_lst_target = np.array([  # noqa
            [35855.23483, 29612.82676],
            [-1025.65590, -1312.08498],
            [6819.32722, -643.00117],
            [27654.86062, -1249.67350]])
        # yapf: enable

        self.assertLessEqual(
            np.linalg.norm(result_lst_target - result_lst_eval),
            1e-4,
            "The optimal values are not returned correctly.",
        )

        self.assertLessEqual(
            np.linalg.norm(H_eig_lst_target - H_eig_lst_eval),
            1e-4,
            "The Hessian values are not returned correctly.",
        )



import unittest
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight
import numpy as np

from manipulation import FindResource


class TestPoseEstimation(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(4)
    @timeout_decorator.timeout(2.0)
    def test_filtered_pointcloud(self):
        """Test filtered pointcloud"""
        # Round to 3 decimal places to avoid floating point sensitivity
        f = np.around(
            self.notebook_locals["scene_pcl_np_filtered"].T, decimals=3
        ).tolist()
        f_target = np.load(FindResource("exercises/pose/pcl_filtered.npy"))
        f_target = np.around(f_target, decimals=3).tolist()

        f = set(tuple(row) for row in f)
        f_target = set(tuple(row) for row in f_target)

        inliers = len(f.intersection(f_target))
        outliers = len(f) - inliers

        self.assertLessEqual(
            outliers,
            80,
            "You have too many false positives (too many outliers)",
        )
        self.assertLessEqual(
            len(f_target) - inliers,
            170,
            "You have too many false negatives (not enough inliers)",
        )



import unittest
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight
import numpy as np


class TestRobotPainter(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(4)
    @timeout_decorator.timeout(1.0)
    def test_key_frames(self):
        """compose_circular_key_frames"""
        f = self.notebook_locals["compose_circular_key_frames"]
        X_WC = self.notebook_locals["X_WCenter"]
        X_WG = self.notebook_locals["X_WG"]
        radius = self.notebook_locals["radius"]
        thetas = self.notebook_locals["thetas"]
        # carry out computation
        output_frames = f(thetas, X_WC, X_WG)

        # test all key positions match radius
        for i, frame_i in enumerate(output_frames):
            if i == 0:
                # check if the first frame is the gripper's current frame
                dist = np.linalg.norm(
                    frame_i.translation() - X_WG.translation()
                )
                self.assertLessEqual(
                    dist, 1e-6, "first key frame position incorrect!"
                )
            else:
                # check if the radius is correct
                pos_cur = frame_i.translation()
                r_cur = np.linalg.norm(pos_cur - X_WC.translation())
                self.assertLessEqual(
                    abs(radius - r_cur),
                    1e-6,
                    "key frame positions incorrect!",
                )

                # check if +z of each frame points toward the center
                z_cur = frame_i.rotation().matrix()[0:3, 2]
                test_center = z_cur * radius + pos_cur
                center_err = np.linalg.norm(test_center - X_WC.translation())
                self.assertLessEqual(
                    center_err, 1e-6, "key frame orientations incorrect!"
                )



from typing import Callable, Optional, Union
import warnings

import gym
import numpy as np

from pydrake.common import RandomGenerator
from pydrake.systems.analysis import Simulator, SimulatorStatus
from pydrake.systems.framework import (
    Context,
    InputPort,
    InputPortIndex,
    OutputPortIndex,
    PortDataType,
    System,
)
from pydrake.systems.sensors import ImageRgba8U


class DrakeGymEnv(gym.Env):
    """
    DrakeGymEnv provides a gym.Env interface for a Drake System (often a
    Diagram) using a Simulator.
    """

    def __init__(
        self,
        simulator: Union[Simulator, Callable[[RandomGenerator], Simulator]],
        time_step: float,
        action_space: gym.spaces.space,
        observation_space: gym.spaces.space,
        reward: Union[
            Callable[[System, Context], float], OutputPortIndex, str
        ],
        action_port_id: Union[InputPort, InputPortIndex, str] = None,
        observation_port_id: Union[OutputPortIndex, str] = None,
        render_rgb_port_id: Union[OutputPortIndex, str] = None,
        set_home: Callable[[Simulator, Context], None] = None,
        hardware: bool = False,
    ):
        """
        Args:
            simulator: Either:
                * A drake.systems.analysis.Simulator, or
                * A function that produces a (randomized) Simulator.
            time_step: Each call to step() will advance the simulator by
                `time_step` seconds.
            reward: The reward can be specified in one of two
                ways: (1) by passing a callable with the signature
                `value = reward(context)` or (2) by passing a scalar
                vector-valued output port of `simulator`'s system.
            action_port_id: The ID of an input port of `simulator`'s system
                compatible with the action_space.  Each Env *must* have an
                action port; passing `None` defaults to using the *first*
                input port (inspired by
                `InputPortSelection.kUseFirstInputIfItExists`).
            action_space: Defines the `gym.spaces.space` for the actions.  If
                the action port is vector-valued, then passing `None` defaults
                to a gym.spaces.Box of the correct dimension with bounds at
                negative and positive infinity.  Note: Stable Baselines 3
                strongly encourages normalizing the action_space to [-1, 1].
            observation_port_id: An output port of `simulator`'s system
                compatible with the observation_space. Each Env *must* have
                an observation port (it seems that gym doesn't support empty
                observation spaces / open-loop policies); passing `None`
                defaults to using the *first* input port (inspired by
                `OutputPortSelection.kUseFirstOutputIfItExists`).
            observation_space: Defines the gym.spaces.space for the
                observations.  If the observation port is vector-valued, then
                passing `None` defaults to a gym.spaces.Box of the correct
                dimension with bounds at negative and positive infinity.
            render_rgb_port: An optional output port of `simulator`'s system
                that returns  an `ImageRgba8U`; often the `color_image` port
                of a Drake `RgbdSensor`.  When not `None`, this enables the
                environment `render_mode` `rgb_array`.
            set_home: A function that sets the home state (plant, and/or env.)
                at reset(). The reset state can be specified in one of
                the two ways:
                (1) setting random context using a Drake random_generator
                (e.g. joint.set_random_pose_distribution()),
                (2) parssing a function set_home().
            hardware: If True, it prevents from setting random context at
                reset() when using random_generator, but it does execute
                set_home() if given.


        Notes (using `env` as an instance of this class):
        - You may set simulator/integrator preferences by using `env.simulator`
          directly.
        - The `done` condition returned by `step()` is always False by
          default.  Use `env.simulator.set_monitor()` to use Drake's monitor
          functionality for specifying termination conditions.
        - You may additionally wish to directly set `env.reward_range` and/or
          `env.spec`.  See the docs for gym.Env for more details.
        """
        super().__init__()
        if isinstance(simulator, Simulator):
            self.simulator = simulator
            self.make_simulator = None
        elif callable(simulator):
            self.simulator = None
            self.make_simulator = simulator
        else:
            raise ValueError("Invalid simulator argument")

        assert time_step > 0
        self.time_step = time_step

        assert isinstance(action_space, gym.spaces.Space)
        self.action_space = action_space

        assert isinstance(observation_space, gym.spaces.Space)
        self.observation_space = observation_space

        if isinstance(reward, (OutputPortIndex, str)):
            self.reward_port_id = reward
            self.reward = None
        elif callable(reward):
            self.reward_port_id = None
            self.reward = reward
        else:
            raise ValueError("Invalid reward argument")

        if action_port_id:
            assert isinstance(action_port_id, (InputPortIndex, str))
            self.action_port_id = action_port_id
        else:
            self.action_port_id = InputPortIndex(0)

        if observation_port_id:
            assert isinstance(observation_port_id, (OutputPortIndex, str))
            self.observation_port_id = observation_port_id
        else:
            self.observation_port_id = OutputPortIndex(0)

        self.metadata["render.modes"] = ["human", "ascii"]

        # (Maybe) setup rendering
        if render_rgb_port_id:
            assert isinstance(render_rgb_port_id, (OutputPortIndex, str))
            self.metadata["render.modes"].append("rgb_array")
        self.render_rgb_port_id = render_rgb_port_id

        self.generator = RandomGenerator()

        if set_home is None or callable(set_home):
            self.set_home = set_home
        else:
            raise ValueError("Invalid set_home argument")

        self.hardware = hardware

        if self.simulator:
            self._setup()

    def _setup(self):
        """Completes the setup once we have a self.simulator."""
        system = self.simulator.get_system()

        # Setup action port
        if self.action_port_id:
            if isinstance(self.action_port_id, InputPortIndex):
                self.action_port = system.get_input_port(self.action_port_id)
            else:
                self.action_port = system.GetInputPort(self.action_port_id)
        if self.action_port.get_data_type() == PortDataType.kVectorValued:
            assert np.array_equal(
                self.action_space.shape, [self.action_port.size()]
            )

        def get_output_port(id):
            if isinstance(id, OutputPortIndex):
                return system.get_output_port(id)
            return system.GetOutputPort(id)

        # Setup observation port
        if self.observation_port_id:
            self.observation_port = get_output_port(self.observation_port_id)
        if self.observation_port.get_data_type() == PortDataType.kVectorValued:
            assert np.array_equal(
                self.observation_space.shape, [self.observation_port.size()]
            )

        # Note: We require that there is no direct feedthrough action_port to
        # observation_port.  Unfortunately, HasDirectFeedthrough returns false
        # positives, and would produce noisy warnings.

        # Setup reward
        if self.reward_port_id:
            reward_port = get_output_port(self.reward_port_id)
            self.reward = lambda system, context: reward_port.Eval(context)[0]

        # (Maybe) setup rendering port
        if self.render_rgb_port_id:
            self.render_rgb_port = get_output_port(self.render_rgb_port_id)
            assert (
                self.render_rgb_port.get_data_type()
                == PortDataType.kAbstractValued
            )
            assert isinstance(
                self.render_rgb_port.Allocate().get_value(), ImageRgba8U
            )

    def step(self, action):
        """
        Implements gym.Env.step to advance the simulation forward by one
        `self.time_step`.

        Args:
            action: an element from self.action_space
        """
        assert self.simulator, "You must call reset() first"

        context = self.simulator.get_context()
        time = context.get_time()

        self.action_port.FixValue(context, action)
        truncated = False
        try:
            status = self.simulator.AdvanceTo(time + self.time_step)
        except RuntimeError as e:
            if (
                "MultibodyPlant's discrete update solver failed to converge"
                not in e.args[0]
            ):
                raise
            warnings.warn("Calling Done after catching RuntimeError:")
            warnings.warn(e.args[0])
            truncated = True
            status = e.args[0]

        observation = self.observation_port.Eval(context)
        reward = self.reward(self.simulator.get_system(), context)
        reached_term = (
            SimulatorStatus.ReturnReason.kReachedTerminationCondition
        )
        terminated = not truncated and (status.reason() == reached_term)
        info = dict()

        # TODO(ggould) This interface is in flux and lags its documentation;
        # when gym is next upgraded it will begin to generate warnings.  At
        # that time replace `(terminated or truncated)`
        # with `terminated, truncated`.
        return observation, reward, (terminated or truncated), info

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[dict] = None
    ):
        """
        If a callable "simulator factory" was passed to the constructor, then a
        new simulator is created.  Otherwise this method simply resets the
        `simulator` and its Context.
        """
        assert options is None  # No options supported yet.

        if seed is not None:
            # TODO(ggould) This should not reset the generator if it was
            # already explicitly seeded (see API spec), but we have no way to
            # check that at the moment.
            self.generator = RandomGenerator(seed)

        if self.make_simulator:
            self.simulator = self.make_simulator(self.generator)
            self._setup()

        context = self.simulator.get_mutable_context()
        context.SetTime(0)
        self.simulator.Initialize()
        if self.set_home is not None:
            self.simulator.get_system().SetDefaultContext(context)
            self.set_home(self.simulator, context, seed)
        else:
            if not self.hardware:
                self.simulator.get_system().SetRandomContext(
                    context, self.generator
                )
        # Note: The output port will be evaluated without fixing the input
        # port.
        observations = self.observation_port.Eval(context)
        return observations if not return_info else (observations, dict())

    def render(self, mode="human"):
        """
        Rendering in `human` mode is accomplished by calling Publish on
        `system`.  This should cause visualizers inside the System (e.g.
        MeshcatVisualizer, PlanarSceneGraphVisualizer, etc.) to draw their
        outputs.  To be fully compliant, those visualizers should set their
        default publishing period to `np.inf` (do not publish periodically).

        Rendering in `ascii` mode calls __repr__ on the system Context.

        Rendering in `rgb_array` mode is enabled by passing a compatible
        `render_rgb_port` to the class constructor.
        """
        assert self.simulator, "You must call reset() first"

        if mode == "human":
            self.simulator.get_system().ForcedPublish(
                self.simulator.get_context()
            )
            return
        elif mode == "ansi":
            return __repr__(self.simulator.get_context())
        elif mode == "rgb_array":
            assert (
                self.render_rgb_port
            ), "You must set render_rgb_port in the constructor"
            return self.render_rgb_port.Eval(
                self.simulator.get_context()
            ).data[:, :, :3]
        else:
            super(DrakeGymEnv, self).render(mode=mode)

    def seed(self, seed=None):
        """Implements gym.Env.seed using Drake's RandomGenerator."""
        if seed:
            self.generator = RandomGenerator(seed)
        else:
            seed = self.generator()
        # Note: One could call self.action_space.seed(self.generator()) here,
        # but it appears that is not the standard approach:
        # https://github.com/openai/gym/issues/681
        return [seed]



import os
import sys
from datetime import date
from pathlib import Path
from urllib.request import urlretrieve
from warnings import warn

import numpy as np
from IPython import get_ipython
from pydrake.all import GetDrakePath
from pydrake.common import GetDrakePath
from pydrake.geometry import RenderLabel

# Use a global variable here because some calls to IPython will actually case an
# interpreter to be created.  This file needs to be imported BEFORE that
# happens.
running_as_notebook = (
    "COLAB_TESTING" not in os.environ
    and get_ipython()
    and hasattr(get_ipython(), "kernel")
)

running_as_test = False


def set_running_as_test(value):
    global running_as_test
    running_as_test = value


def pyplot_is_interactive():
    # import needs to happen after the backend is set.
    import matplotlib.pyplot as plt
    from matplotlib.rcsetup import interactive_bk

    return plt.get_backend() in interactive_bk


def FindResource(filename):
    return os.path.join(os.path.dirname(__file__), filename)


# A filename from the data directory.  Will download if necessary.
def LoadDataResource(filename):
    data = os.path.join(os.path.dirname(os.path.dirname(__file__)), "data")
    if not os.path.exists(data):
        os.makedirs(data)
    path = os.path.join(data, filename)
    if not os.path.exists(path):
        print(f"{path} was not found locally; downloading it now...")
        urlretrieve(
            f"https://manipulation.csail.mit.edu/data/{filename}", path
        )
    return path


def ConfigureParser(parser):
    """Add the manipulation/package.xml index to the given Parser."""
    package_xml = os.path.join(os.path.dirname(__file__), "models/package.xml")
    parser.package_map().AddPackageXml(filename=package_xml)
    AddPackagePaths(parser)


def AddPackagePaths(parser):
    # Remove once https://github.com/RobotLocomotion/drake/issues/10531 lands.
    parser.package_map().PopulateFromFolder(FindResource(""))
    parser.package_map().Add(
        "manipulation_station",
        os.path.join(
            GetDrakePath(),
            "examples/manipulation_station/models",
        ),
    )
    parser.package_map().Add(
        "ycb",
        os.path.join(GetDrakePath(), "manipulation/models/ycb"),
    )
    parser.package_map().Add(
        "wsg_50_description",
        os.path.join(
            GetDrakePath(),
            "manipulation/models/wsg_50_description",
        ),
    )


reserved_labels = [
    RenderLabel.kDoNotRender,
    RenderLabel.kDontCare,
    RenderLabel.kEmpty,
    RenderLabel.kUnspecified,
]


def colorize_labels(image):
    # import needs to happen after backend is set up.
    import matplotlib as mpl
    import matplotlib.pyplot as plt

    """Colorizes labels."""
    # TODO(eric.cousineau): Revive and use Kuni's palette.
    cc = mpl.colors.ColorConverter()
    color_cycle = plt.rcParams["axes.prop_cycle"]
    colors = np.array([cc.to_rgb(c["color"]) for c in color_cycle])
    bg_color = [0, 0, 0]
    image = np.squeeze(image)
    background = np.zeros(image.shape[:2], dtype=bool)
    for label in reserved_labels:
        background |= image == int(label)
    image[np.logical_not(background)]
    color_image = colors[image % len(colors)]
    color_image[background] = bg_color
    return color_image


def SetupMatplotlibBackend(wishlist=["notebook"]):
    """
    Helper to support multiple workflows:
        1) nominal -- running locally w/ jupyter notebook
        2) unit tests (no ipython, backend is template)
        3) binder -- does have notebook backend
        4) colab -- claims to have notebook, but it doesn't work
    Puts the matplotlib backend into notebook mode, if possible,
    otherwise falls back to inline mode.
    Returns True iff the final backend is interactive.
    """
    # To find available backends, one can access the lists:
    # matplotlib.rcsetup.interactive_bk
    # matplotlib.rcsetup.non_interactive_bk
    # matplotlib.rcsetup.all_backends
    if running_as_notebook:
        ipython = get_ipython()
        # Short-circuit for google colab.
        if "google.colab" in sys.modules:
            ipython.run_line_magic("matplotlib", "inline")
            return False
        # TODO: Find a way to detect vscode, and use inline instead of notebook
        for backend in wishlist:
            try:
                ipython.run_line_magic("matplotlib", backend)
                return pyplot_is_interactive()
            except KeyError:
                continue
        ipython.run_line_magic("matplotlib", "inline")
    return False


def DrakeVersionGreaterThan(minimum_date: date):
    drake_version_txt = (
        Path(GetDrakePath()).parent / "doc" / "drake" / "VERSION.TXT"
    )
    # If the file doesn't exist, then we should pass. A source install won't
    # have VERSION.TXT
    if drake_version_txt.is_file():
        with open(drake_version_txt, "r") as f:
            drake_version = f.read()
        drake_version = drake_version.split()[0]
        if len(drake_version) == 14:
            drake_date = date(
                year=int(drake_version[:4]),
                month=int(drake_version[4:6]),
                day=int(drake_version[6:8]),
            )
        elif drake_version == "1.13.0":
            drake_date = date(year=2023, month=2, day=14)
        elif drake_version == "1.14.0":
            drake_date = date(year=2023, month=3, day=15)
        else:
            warn(f"Unrecognized drake version {drake_version}")
            return
        if drake_date < minimum_date:
            raise (
                RuntimeError(
                    f"You need to update your Drake version. Python is using the Drake installation in {GetDrakePath()}. This installation was from a nightly build on {drake_date}, but this method requires Drake from at least {minimum_date}."
                )
            )


# Adapted from Drake's system_doxygen.py.  Just make an html rendering of the
# system block with its name and input/output ports (even if it is a Diagram).
def SystemHtml(system):
    input_port_html = ""
    for p in range(system.num_input_ports()):
        input_port_html += (
            f'<tr><td align=right style="padding:5px 0px 5px 0px">'
            f"{system.get_input_port(p).get_name()} &rarr;</td></tr>"
        )
    output_port_html = ""
    for p in range(system.num_output_ports()):
        output_port_html += (
            '<tr><td align=left style="padding:5px 0px 5px 0px">'
            f"&rarr; {system.get_output_port(p).get_name()}</td></tr>"
        )
    # Note: keeping this on a single line avoids having to handle comment line
    # markers (e.g. * or ///)
    html = (
        f"<table align=center cellpadding=0 cellspacing=0><tr align=center>"
        f'<td style="vertical-align:middle">'
        f"<table cellspacing=0 cellpadding=0>{input_port_html}</table>"
        f"</td>"
        f'<td align=center style="border:2px solid black;padding-left:20px;'
        f'padding-right:20px;vertical-align:middle" bgcolor=#F0F0F0>'
        f"{system.get_name()}</td>"
        f'<td style="vertical-align:middle">'
        f"<table cellspacing=0 cellpadding=0>{output_port_html}</table>"
        f"</td></tr></table>"
    )

    return html



import unittest
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight
import numpy as np
from pydrake.all import RotationMatrix


class TestSimulationTuning(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @timeout_decorator.timeout(2.0)
    @weight(2)
    def test_on_slope(self):
        """Test test_on_slope"""
        simulator = self.notebook_locals["simulator_d"]
        diagram = self.notebook_locals["diagram_d"]

        context = simulator.get_context()
        plant = diagram.GetSubsystemByName("plant")
        plant_context = plant.GetMyMutableContextFromRoot(context)
        box_poses = plant.GetPositions(plant_context)
        box1_pos = box_poses[:7][4:]
        box2_pos = box_poses[7:][4:]

        box_pos_range = np.array(
            [
                [0.0, 1.0],
                [-0.1, 0.1],
                [0.03, 0.07],
            ]
        )

        in_range_list1 = []
        in_range_list2 = []
        for i in range(3):
            in_range_1 = (
                box_pos_range[i, 0] < box1_pos[i] < box_pos_range[i, 1]
            )
            in_range_2 = (
                box_pos_range[i, 0] < box2_pos[i] < box_pos_range[i, 1]
            )
            in_range_list1.append(in_range_1)
            in_range_list2.append(in_range_2)

        in_range_all = np.asarray(in_range_list1 + in_range_list2)
        self.assertTrue(
            in_range_all.all(), "Final box positions not in correct range!"
        )

    @timeout_decorator.timeout(2.0)
    @weight(2)
    def test_make_simulation(self):
        """Test test_make_simulation"""
        simulator = self.notebook_locals["simulator_f"]
        diagram = self.notebook_locals["diagram_f"]

        context = simulator.get_context()
        plant = diagram.GetSubsystemByName("plant")
        plant_context = plant.GetMyMutableContextFromRoot(context)
        box_poses = plant.GetPositions(plant_context)
        box1_pos = box_poses[:7][4:]
        box2_pos = box_poses[7:][4:]

        # box1_frame = plant.GetBodyByName('box').body_frame()
        # box2_frame = plant.GetBodyByName('box_2').body_frame()

        # box1_pose = box1_frame.CalcPoseInWorld(plant_context).GetAsMatrix4()
        # box2_pose = box2_frame.CalcPoseInWorld(plant_context).GetAsMatrix4()
        # print(f'Box 1 pose: {box1_pose}')
        # print(f'Box 2 pose: {box2_pose}')

        box_pos_range = np.array(
            [
                [0.0, 1.0],
                [-0.1, 0.1],
                [0.03, 0.07],
            ]
        )

        in_range_list1 = []
        in_range_list2 = []
        for i in range(3):
            in_range_1 = (
                box_pos_range[i, 0] < box1_pos[i] < box_pos_range[i, 1]
            )
            in_range_2 = (
                box_pos_range[i, 0] < box2_pos[i] < box_pos_range[i, 1]
            )
            in_range_list1.append(in_range_1)
            in_range_list2.append(in_range_2)

        in_range_all = np.asarray(in_range_list1 + in_range_list2)
        self.assertTrue(
            in_range_all.all(), "Final box positions not in correct range!"
        )

    @timeout_decorator.timeout(2.0)
    @weight(3)
    def test_make_stacking_simulation(self):
        """Test test_make_stacking_simulation"""
        simulator = self.notebook_locals["simulator_h"]
        diagram = self.notebook_locals["diagram_h"]

        context = simulator.get_context()
        plant = diagram.GetSubsystemByName("plant")
        plant_context = plant.GetMyMutableContextFromRoot(context)
        box_poses = plant.GetPositions(plant_context)
        box1_pos = box_poses[:7][4:]
        box2_pos = box_poses[7:][4:]

        # box1_frame = plant.GetBodyByName('box').body_frame()
        # box2_frame = plant.GetBodyByName('box_2').body_frame()

        # box1_pose = box1_frame.CalcPoseInWorld(plant_context).GetAsMatrix4()
        # box2_pose = box2_frame.CalcPoseInWorld(plant_context).GetAsMatrix4()
        # print(f'Box 1 pose: {box1_pose}')
        # print(f'Box 2 pose: {box2_pose}')

        box1_pos_range = np.array(
            [[0.2, 0.3], [-0.025, 0.025], [0.025, 0.0375]]
        )
        box2_pos_range = np.array(
            [[0.2, 0.3], [-0.025, 0.025], [0.0475, 0.06]]
        )

        in_range_list1 = []
        in_range_list2 = []
        for i in range(3):
            in_range_1 = (
                box1_pos_range[i, 0] < box1_pos[i] < box1_pos_range[i, 1]
            )
            in_range_2 = (
                box2_pos_range[i, 0] < box2_pos[i] < box2_pos_range[i, 1]
            )
            in_range_list1.append(in_range_1)
            in_range_list2.append(in_range_2)

        in_range_all = np.asarray(in_range_list1 + in_range_list2)
        self.assertTrue(
            in_range_all.all(), "Final box positions not in correct range!"
        )

    @timeout_decorator.timeout(2.0)
    @weight(2)
    def test_matching_coll_shape(self):
        """Test test_matching_coll_shape"""
        plant = self.notebook_locals["plant_a"]
        diagram = self.notebook_locals["diagram_a"]

        context = diagram.CreateDefaultContext()
        plant_context = plant.GetMyContextFromRoot(context)
        contact_results = plant.get_contact_results_output_port().Eval(
            plant_context
        )

        box_q_list = np.array(
            [
                [0.0, 0.15, 0.0],
                [0.0, 0.2, 0.0],
                [0.0, 0.2, 0.0],
                [0.0, 0.13, 0.0],
            ]
        )

        box2_q_list = np.array(
            [
                [-0.049, 0.198, 0.0],
                [-0.049, 0.155, 0.0],
                [0.046, 0.168, 0.0],
                [0.046, 0.168, 0.0],
            ]
        )

        n_contacts_list = []
        for i in range(4):
            plant.SetPositions(
                plant_context, np.concatenate([box_q_list[i], box2_q_list[i]])
            )

            contact_results = plant.get_contact_results_output_port().Eval(
                plant_context
            )
            n_contacts = contact_results.num_point_pair_contacts()

            n_contacts_list.append(n_contacts)

        self.assertTrue(
            (np.array(n_contacts) > 0).all(), "Objects not contacting!"
        )

    @timeout_decorator.timeout(2.0)
    @weight(2)
    def test_force_discontinuity(self):
        """Test test_force_discontinuity"""
        plant = self.notebook_locals["plant_a"]
        diagram = self.notebook_locals["diagram_a"]
        f = self.notebook_locals["set_block_2d_poses"]

        box1_pos1, box2_pos1, box1_pos2, box2_pos2 = f()

        context = diagram.CreateDefaultContext()
        plant_context = plant.GetMyContextFromRoot(context)

        plant.SetPositions(
            plant_context, np.concatenate([box1_pos1 + box2_pos1])
        )

        plant.get_contact_results_output_port()
        contact_results = plant.get_contact_results_output_port().Eval(
            plant_context
        )
        info = contact_results.point_pair_contact_info(0)
        force_1 = info.contact_force()
        fn1 = force_1 / np.linalg.norm(force_1)
        point_1 = info.contact_point()

        plant.SetPositions(
            plant_context, np.concatenate([box1_pos2 + box2_pos2])
        )

        plant.get_contact_results_output_port()
        contact_results = plant.get_contact_results_output_port().Eval(
            plant_context
        )
        info = contact_results.point_pair_contact_info(0)
        force_2 = info.contact_force()
        fn2 = force_2 / np.linalg.norm(force_2)
        point_2 = info.contact_point()

        contact_pt_dist = np.linalg.norm(
            np.array([point_1[0], point_1[2]])
            - np.array([point_2[0], point_2[2]])
        )
        force_angle = np.abs(np.arccos(np.dot(fn1, fn2)))

        self.assertLessEqual(
            contact_pt_dist,
            0.01,
            "Contact points not close enough, discontinuity not detected",
        )

        self.assertGreaterEqual(
            force_angle,
            np.deg2rad(60),
            "Angle between force vectors too small, discontinuity not detected",
        )

    @timeout_decorator.timeout(2.0)
    @weight(2)
    def test_multi_contact(self):
        """Test test_multi_contact"""
        simulator = self.notebook_locals["simulator_e"]
        diagram = self.notebook_locals["diagram_e"]

        context = simulator.get_context()
        plant = diagram.GetSubsystemByName("plant")
        plant_context = plant.GetMyMutableContextFromRoot(context)

        plant.get_contact_results_output_port()
        contact_results = plant.get_contact_results_output_port().Eval(
            plant_context
        )
        n_contacts = contact_results.num_point_pair_contacts()

        self.assertGreater(n_contacts, 2, "Not enough contacts detected!")

    @timeout_decorator.timeout(5.0)
    @weight(2)
    def test_minimal_rotation(self):
        """Test test_minimal_rotation"""
        sim_maker = self.notebook_locals["sim_maker"]
        f = sim_maker.make_simulation
        simulator, diagram = f(0.001, (1, 1), (0.1, 0.1))

        context = simulator.get_context()
        plant = diagram.GetSubsystemByName("plant")
        plant_context = plant.GetMyMutableContextFromRoot(context)

        box2_frame = plant.GetBodyByName("box_2").body_frame()

        final_rot = box2_frame.CalcPoseInWorld(plant_context).rotation()
        expected_rot = RotationMatrix.MakeYRotation(0.1 + np.pi / 2)

        rot_diff = final_rot.inverse() @ expected_rot
        diff_norm = np.linalg.norm(rot_diff.matrix() - np.eye(3))

        self.assertLessEqual(diff_norm, 0.1, "Box 2 moved more than expected!")
